# Analysis of Hidden Markov Model Learning Algorithms for the Detection and Prediction of Multi-stage Network Attacks
This code reproduces the results of the journal article titled **Analysis of Hidden Markov Model Learning Algorithms for the Detection and Prediction of Multi-stage Network Attacks** by Timothy Chadza, Konstantinos Kyriakopoulos & Sangarapillai Lambotharan published in Elsevier's Future Generation Systems Journal, https://doi.org/10.1016/j.future.2020.03.014. If you use this code, please, cite the above paper.

To run this code, simply, run the `main.m` function. 

The experiment parameters can be changed by editing the `main.m` file. When the experiments are run, the results are observed in the following figures:
* `fig1_AS.png`: detection of all states (AS);
* `fig2_CS.png`: detection of current state (CS);
* `fig3_NS1_Method1.png`: prediction of next state (NS) level 1 using method 1 (A, B and alpha);
* `fig4_NS2_Method1.png`: prediction of NS level 2 using method 1 (A, B and alpha);
* `fig5_NO1_Method1.png`: prediction of next observation (NO) level 1 using method 1 (A, B and alpha);
* `fig6_NO2_Method1.png`: prediction of NO level 2 using method 1 (A, B and alpha);
* `fig7_NS1_Method2.png`: prediction of NS level 1 using method 2 (A and alpha);
* `fig8_NS1_Method2.png`: prediction of NS level 2 using method 2 (A and alpha);
* `fig9_NO2_Method2.png`: prediction of NO level 1 using method 2 (B and alpha); 
* `fig10_NO2_Method2.png`: prediction of NO level 2 using method 2 (B and alpha). 

## Included Materials

In addition to this README.md, the following files are included in this code:

* `accuracy.m`: computes the detection and prediction accuracy. The AS and CS use Viterbi decoding, whereas the developed NS and NO functions are called for prediction computation.
* `count_based.m`: obtains the count-based parameters for HMM initialisation.
* `data_extract.m`: extracts the fields from Snort alerts. This can be skipped if alerts are already in CSV format i.e obtained from Barnyard2 and loaded in MySQL database.
* `data_time.m`:  converts the timestamp to MATLAB format. Requires financial toolbox to be installed for standalone implementation.
* `differential_evolution.m`: implements the differential evolution algorithm.
* `evaluate.m`: evaluates the detection and prediction accuracy of HMMs. It calls the accuracy.m function for computation.
* `initialise_random.m`: determines randomly initialised HMM parameters, A, B and Pi.  Default parameters are provided in `ABPi.mat` but this function can be enabled in `main.m`.
* `k_means_default.m`: computes the B parameter using K-means. The assumption is that all the alerts are considered regardless of their phases. K-means is then applied to the dataset to determine the K-clusters for the observation sequences where K=N\*M. Thereafter, these clusters are used to estimate B. 
* `k_means_phases.m`: computes the B parameter using K-means. This approach assummes that all the alerts are considered based on the phases. Hence N phases are used as a starting point and K-means is applied on the phases, where K=M, thereafter, the clusters are used to compute B.
* `learnAHK.m`: a spectral algorithm implementation developed by Carl Mattfield and can be cited using,  C. Mattfield, "Implementing Spectral Methods for Hidden Markov Models with Real-Valued Emissions",arXiv:1404.7472, 2014". It implements the algorithm defined by Anandkumar, Animashree, Daniel Hsu, and Sham M. Kakade. "A method of moments for mixture models and hidden Markov models." arXiv preprint arXiv:1203.0683 (2012). Two lines of code are appended to determine the HMM parameters. 'AHK' is used to represent this algorithm.
* `learnHMM.m`: a spectral algorithm developed by Carl Mattfield and cites the spectral algorithm concepts defined by Hsu, Daniel, Sham M. Kakade, and Tong Zhang. "A spectral algorithm for learning hidden Markov models." Journal of Computer and System Sciences 78.5 (2012): 1460-1480. Two lines of code are appended to determine the HMM parameters. 'HKZ' is used to represent this algorithm.
* `main.m`: main file implementing all the experiments.
* `next_state_observ_Aalpha_Balpha.m`: computes the next state and next observation for two levels using method 2.
* `next_state_observ_ABalpha.m`: computes the next state and next observation for two levels using method 1.
* `normalise_rows.m`: assigns small values to zeroes and normalises the rows of HMM parameters so that each row sums to unity. Also, if negative parameters are obtained, they are scaled to start from zero and then normalised.
* `number2month.m`: converts digit months to words into suitable MATLAB format for easier processing. The extracted timestamp is in a different format than the one used by MATLAB.
* `plottingresults.m`: plots the results. Only the sequential sampling results for both CML and TL are specified but this can be modified directly in this function.
* `preprocess.m`: formats data into cells while extracting fields from raw data, removes duplicate alerts, modifies the time to MATLAB format, obtains the relative time, and assigns state and observation symbols.
* `remove_dup.m`: remove duplicate alerts. 
* `seq_triple`: splits the observation sequence into N-tuples for the spectral algorithm deployment. The code was supplemented by Carl Mattfield as an additional concept to implement HMM on spectral algorithm.
* `string_num`: converts string to a number.
* `symbols_observations`: assigns symbols to observations.
* `symbols_states_DMZ1`: appends state symbols to alerts in the DMZ dataset. 
* `symbols_states_inside1`:  appends state symbols to alerts in the inside dataset. 
* `ABPi.mat`: the HMM randomly initialised parameters for replication. 
* `alert_1dmz2949`: raw Snort IDS alerts for DARPA 2000 scenario one-DMZ.
* `alert_1inside2306`: raw Snort IDS alerts for DARPA 2000 scenario one-inside network.

### Authors
Timothy Chadza, Konstantinos Kyriakopoulos & Sangarapillai Lambotharan 

Wolfson School of Mechanical, Electrical and Manufacturing Engineering,
Loughborough University, Loughborough, UK
